@@ -1,7 +1,5 @@
 /* 
-   arch-tag: Implementation of Rhythmbox playlist parser
-
-   Copyright (C) 2002, 2003, 2004, 2005 Bastien Nocera
+   Copyright (C) 2002, 2003, 2004, 2005, 2006 Bastien Nocera
    Copyright (C) 2003, 2004 Colin Walters <walters@rhythmbox.org>
 
    The Gnome Library is free software; you can redistribute it and/or
@@ -26,21 +24,24 @@
 #include "config.h"
 #endif /* HAVE_CONFIG_H */
 
-#include "totem-pl-parser.h"
-
-#include "totemplparser-marshal.h"
-#include "totem-disc.h"
-
+#include <string.h>
 #include <glib.h>
-#include <glib/gi18n.h>
-#include <gtk/gtk.h>
+#include <glib/gi18n-lib.h>
+#include <libgnomevfs/gnome-vfs-mime-utils.h>
+
+#ifndef TOTEM_PL_PARSER_MINI
 #include <libxml/tree.h>
 #include <libxml/parser.h>
+#include <gtk/gtk.h>
 #include <libgnomevfs/gnome-vfs.h>
 #include <libgnomevfs/gnome-vfs-mime.h>
-#include <libgnomevfs/gnome-vfs-mime-utils.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
-#include <string.h>
+#include "totem-pl-parser.h"
+#include "totemplparser-marshal.h"
+#include "totem-disc.h"
+#endif /* !TOTEM_PL_PARSER_MINI */
+
+#include "totem-pl-parser-mini.h"
 
 #define READ_CHUNK_SIZE 8192
 #define MIME_READ_CHUNK_SIZE 1024
@@ -48,14 +49,43 @@
 #define DIR_MIME_TYPE "x-directory/normal"
 #define BLOCK_DEVICE_TYPE "x-special/device-block"
 #define EXTINF "#EXTINF:"
-#define DEBUG(x) { if (parser->priv->debug != FALSE) x; }
+#define DEBUG(x) { if (parser->priv->debug) x; }
 
+typedef gboolean (*PlaylistIdenCallback) (const char *data, gsize len);
+
+#ifndef TOTEM_PL_PARSER_MINI
 typedef TotemPlParserResult (*PlaylistCallback) (TotemPlParser *parser, const char *url, gpointer data);
+#endif
+
+typedef struct {
+	char *mimetype;
+#ifndef TOTEM_PL_PARSER_MINI
+	PlaylistCallback func;
+#endif
+	PlaylistIdenCallback iden;
+} PlaylistTypes;
+
+static gboolean totem_pl_parser_is_ra (const char *data, gsize len);
+static gboolean totem_pl_parser_is_asf (const char *data, gsize len);
+static gboolean totem_pl_parser_is_quicktime (const char *data, gsize len);
+
+#ifndef TOTEM_PL_PARSER_MINI
+
 static gboolean totem_pl_parser_scheme_is_ignored (TotemPlParser *parser, const char *url);
 static gboolean totem_pl_parser_ignore (TotemPlParser *parser, const char *url);
 static TotemPlParserResult totem_pl_parser_parse_internal (TotemPlParser *parser, const char *url);
 static TotemPlParserResult totem_pl_parser_add_asx (TotemPlParser *parser, const char *url, gpointer data);
 
+struct TotemPlParserPrivate
+{
+	GList *ignore_schemes;
+	GList *ignore_mimetypes;
+	guint recurse_level;
+	guint fallback : 1;
+	guint recurse : 1;
+	guint debug : 1;
+};
+
 static void totem_pl_parser_set_property (GObject *object,
 					  guint prop_id,
 					  const GValue *value,
@@ -65,21 +95,6 @@ static void totem_pl_parser_get_property (GObject *object,
 					  GValue *value,
 					  GParamSpec *pspec);
 
-typedef struct {
-	char *mimetype;
-	PlaylistCallback func;
-} PlaylistTypes;
-
-struct TotemPlParserPrivate
-{
-	GList *ignore_schemes;
-	GList *ignore_mimetypes;
-	guint recurse_level;
-	gboolean fallback;
-	gboolean recurse;
-	gboolean debug;
-};
-
 enum {
 	PROP_NONE,
 	PROP_RECURSE,
@@ -94,9 +109,8 @@ enum {
 	LAST_SIGNAL
 };
 
-static int totem_pl_parser_table_signals[LAST_SIGNAL] = { 0 };
-
-static GObjectClass *parent_class = NULL;
+static int totem_pl_parser_table_signals[LAST_SIGNAL];
+static gboolean i18n_done = FALSE;
 
 static void totem_pl_parser_class_init (TotemPlParserClass *class);
 static void totem_pl_parser_init       (TotemPlParser      *parser);
@@ -109,8 +123,6 @@ totem_pl_parser_class_init (TotemPlParserClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
-	parent_class = g_type_class_peek_parent (klass);
-
 	object_class->finalize = totem_pl_parser_finalize;
 	object_class->set_property = totem_pl_parser_set_property;
 	object_class->get_property = totem_pl_parser_get_property;
@@ -170,10 +182,10 @@ totem_pl_parser_set_property (GObject *object,
 	switch (prop_id)
 	{
 	case PROP_RECURSE:
-		parser->priv->recurse = g_value_get_boolean (value);
+		parser->priv->recurse = g_value_get_boolean (value) != FALSE;
 		break;
 	case PROP_DEBUG:
-		parser->priv->debug = g_value_get_boolean (value);
+		parser->priv->debug = g_value_get_boolean (value) != FALSE;
 		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -213,9 +225,21 @@ totem_pl_parser_error_quark (void)
 	return quark;
 }
 
+static void
+totem_pl_parser_init_i18n (void)
+{
+	if (i18n_done == FALSE) {
+		/* set up translation catalog */
+		bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
+		bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
+		i18n_done = TRUE;
+	}
+}
+
 TotemPlParser *
 totem_pl_parser_new (void)
 {
+	totem_pl_parser_init_i18n ();
 	return TOTEM_PL_PARSER (g_object_new (TOTEM_TYPE_PL_PARSER, NULL));
 }
 
@@ -254,6 +278,7 @@ my_gnome_vfs_get_mime_type_with_data (const char *uri, gpointer *data, TotemPlPa
 	/* Read the whole thing, up to MIME_READ_CHUNK_SIZE */
 	buffer = NULL;
 	total_bytes_read = 0;
+	bytes_read = 0;
 	do {
 		buffer = g_realloc (buffer, total_bytes_read
 				+ MIME_READ_CHUNK_SIZE);
@@ -373,12 +398,11 @@ totem_pl_parser_num_entries (TotemPlParser *parser, GtkTreeModel *model,
 	for (i = 1; i <= num_entries; i++)
 	{
 		GtkTreeIter iter;
-		char *path, *url, *title;
+		char *url, *title;
 		gboolean custom_title;
 
-		path = g_strdup_printf ("%d", i - 1);
-		gtk_tree_model_get_iter_from_string (model, &iter, path);
-		g_free (path);
+		if (gtk_tree_model_iter_nth_child (model, &iter, NULL, i -1) == FALSE)
+			return i - ignored;
 
 		func (model, &iter, &url, &title, &custom_title, user_data);
 		if (totem_pl_parser_scheme_is_ignored (parser, url) != FALSE)
@@ -432,6 +456,7 @@ totem_pl_parser_relative (const char *url, const char *output)
 	return base;
 }
 
+#ifndef TOTEM_PL_PARSER_MINI
 static gboolean
 totem_pl_parser_write_pls (TotemPlParser *parser, GtkTreeModel *model,
 			   TotemPlParserIterFunc func, 
@@ -493,12 +518,11 @@ totem_pl_parser_write_pls (TotemPlParser *parser, GtkTreeModel *model,
 
 	for (i = 1; i <= num_entries_total; i++) {
 		GtkTreeIter iter;
-		char *path, *url, *title, *relative;
+		char *url, *title, *relative;
 		gboolean custom_title;
 
-		path = g_strdup_printf ("%d", i - 1);
-		gtk_tree_model_get_iter_from_string (model, &iter, path);
-		g_free (path);
+		if (gtk_tree_model_iter_nth_child (model, &iter, NULL, i - 1) == FALSE)
+			continue;
 
 		func (model, &iter, &url, &title, &custom_title, user_data);
 
@@ -613,12 +637,11 @@ totem_pl_parser_write_m3u (TotemPlParser *parser, GtkTreeModel *model,
 
 	for (i = 1; i <= num_entries_total; i++) {
 		GtkTreeIter iter;
-		char *path, *url, *title, *path2;
+		char *url, *title, *path2;
 		gboolean custom_title;
 
-		path = g_strdup_printf ("%d", i - 1);
-		gtk_tree_model_get_iter_from_string (model, &iter, path);
-		g_free (path);
+		if (gtk_tree_model_iter_nth_child (model, &iter, NULL, i - 1) == FALSE)
+			continue;
 
 		func (model, &iter, &url, &title, &custom_title, user_data);
 
@@ -673,6 +696,108 @@ totem_pl_parser_write_m3u (TotemPlParser *parser, GtkTreeModel *model,
 	return TRUE;
 }
 
+static gboolean
+totem_pl_parser_write_xspf (TotemPlParser *parser, GtkTreeModel *model,
+			   TotemPlParserIterFunc func, 
+			   const char *output, const char *title,
+			   gpointer user_data, GError **error)
+{
+	GnomeVFSHandle *handle;
+	GnomeVFSResult res;
+	int num_entries_total, num_entries, i;
+	char *buf;
+	gboolean success;
+
+	num_entries = totem_pl_parser_num_entries (parser, model, func, user_data);
+	num_entries_total = gtk_tree_model_iter_n_children (model, NULL);
+
+	res = gnome_vfs_open (&handle, output, GNOME_VFS_OPEN_WRITE);
+	if (res == GNOME_VFS_ERROR_NOT_FOUND) {
+		res = gnome_vfs_create (&handle, output,
+				GNOME_VFS_OPEN_WRITE, FALSE,
+				GNOME_VFS_PERM_USER_WRITE
+				| GNOME_VFS_PERM_USER_READ
+				| GNOME_VFS_PERM_GROUP_READ);
+	}
+
+	if (res != GNOME_VFS_OK) {
+		g_set_error(error,
+			    TOTEM_PL_PARSER_ERROR,
+			    TOTEM_PL_PARSER_ERROR_VFS_OPEN,
+			    _("Couldn't open file '%s': %s"),
+			    output, gnome_vfs_result_to_string (res));
+		return FALSE;
+	}
+
+	buf = g_strdup_printf ("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
+				"<playlist version=\"1\" xmlns=\"http://xspf.org/ns/0/\">\n"
+				" <trackList>\n");
+	success = write_string (handle, buf, error);
+	g_free (buf);
+	if (success == FALSE)
+	{
+		gnome_vfs_close (handle);
+		return FALSE;
+	}
+
+	for (i = 1; i <= num_entries_total; i++) {
+		GtkTreeIter iter;
+		char *url, *url_escaped, *relative, *title;
+		gboolean custom_title;
+
+		if (gtk_tree_model_iter_nth_child (model, &iter, NULL, i - 1) == FALSE)
+			continue;
+
+		func (model, &iter, &url, &title, &custom_title, user_data);
+
+		if (totem_pl_parser_scheme_is_ignored (parser, url) != FALSE)
+		{
+			g_free (url);
+			g_free (title);
+			continue;
+		}
+
+		relative = totem_pl_parser_relative (url, output);
+		url_escaped = g_markup_escape_text (relative ? relative : url, -1);
+		buf = g_strdup_printf ("  <track>\n"
+					"   <location>%s</location>\n", url_escaped);
+		success = write_string (handle, buf, error);
+		g_free (url);
+		g_free (url_escaped);
+		g_free (relative);
+		g_free (buf);
+		if (success == FALSE)
+		{
+			gnome_vfs_close (handle);
+			g_free (title);
+			return FALSE;
+		}
+
+		if (custom_title == TRUE)
+			buf = g_strdup_printf ("   <title>%s</title>\n"
+						"  </track>\n", title);
+		else
+			buf = g_strdup_printf ("  </track>\n");
+		
+		success = write_string (handle, buf, error);
+		g_free (buf);
+		g_free (title);
+		if (success == FALSE)
+		{
+			gnome_vfs_close (handle);
+			return FALSE;
+		}
+	}
+
+	buf = g_strdup_printf (" </trackList>\n"
+				"</playlist>");
+	success = write_string (handle, buf, error);
+	g_free (buf);
+	gnome_vfs_close (handle);
+
+	return success;
+}
+
 gboolean
 totem_pl_parser_write_with_title (TotemPlParser *parser, GtkTreeModel *model,
 				  TotemPlParserIterFunc func,
@@ -690,6 +815,9 @@ totem_pl_parser_write_with_title (TotemPlParser *parser, GtkTreeModel *model,
 		return totem_pl_parser_write_m3u (parser, model, func,
 				output, (type == TOTEM_PL_PARSER_M3U_DOS),
                                 user_data, error);
+	case TOTEM_PL_PARSER_XSPF:
+		return totem_pl_parser_write_xspf (parser, model, func,
+				output, title, user_data, error);
 	default:
 		g_assert_not_reached ();
 	}
@@ -708,6 +836,8 @@ totem_pl_parser_write (TotemPlParser *parser, GtkTreeModel *model,
 			NULL, type, user_data, error);
 }
 
+#endif /* TOTEM_PL_PARSER_MINI */
+
 static int
 read_ini_line_int (char **lines, const char *key)
 {
@@ -802,9 +932,7 @@ totem_pl_parser_finalize (GObject *object)
 	g_free (parser->priv);
 	parser->priv = NULL;
 
-	if (G_OBJECT_CLASS (parent_class)->finalize != NULL) {
-		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
-	}
+	G_OBJECT_CLASS (totem_pl_parser_parent_class)->finalize (object);
 }
 
 static gboolean
@@ -832,6 +960,17 @@ totem_pl_parser_add_one_url_ext (TotemPlParser *parser, const char *url,
 		       genre);
 }
 
+static gchar*
+totem_pl_resolve_url (char *base, char *url)
+{
+	if (g_strrstr (url, "://") == NULL) {
+		if (url[0] != '/' || g_str_has_prefix (base, "file://") == FALSE) {
+			return g_strdup_printf ("%s/%s", base, url);
+		}
+	}
+	return g_strdup (url);
+}
+
 static TotemPlParserResult
 totem_pl_parser_add_ram (TotemPlParser *parser, const char *url, gpointer data)
 {
@@ -896,7 +1035,6 @@ static TotemPlParserResult
 totem_pl_parser_add_asf_reference_parser (TotemPlParser *parser,
 		const char *url, gpointer data)
 {
-	gboolean retval = TOTEM_PL_PARSER_RESULT_UNHANDLED;
 	char *contents, **lines, *ref, *split_char;
 	int size;
 
@@ -913,8 +1051,8 @@ totem_pl_parser_add_asf_reference_parser (TotemPlParser *parser,
 	lines = g_strsplit (contents, split_char, 0);
 	g_free (contents);
 
+	/* Try to get Ref1 first */
 	ref = read_ini_line_string (lines, "Ref1", FALSE);
-
 	if (ref == NULL) {
 		g_strfreev (lines);
 		return totem_pl_parser_add_asx (parser, url, data);
@@ -925,12 +1063,14 @@ totem_pl_parser_add_asf_reference_parser (TotemPlParser *parser,
 		memcpy(ref, "mmsh", 4);
 
 	totem_pl_parser_add_one_url (parser, ref, NULL);
-	retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
 	g_free (ref);
 
+	/* Don't try to get Ref2, as it's only ever
+	 * supposed to be a fallback */
+
 	g_strfreev (lines);
 
-	return retval;
+	return TOTEM_PL_PARSER_RESULT_SUCCESS;
 }
 
 static TotemPlParserResult
@@ -1185,8 +1325,11 @@ totem_pl_parser_add_m3u (TotemPlParser *parser, const char *url, gpointer data)
 		/* Either it's a URI, or it has a proper path ... */
 		if (strstr(lines[i], "://") != NULL
 				|| lines[i][0] == G_DIR_SEPARATOR) {
-			totem_pl_parser_add_one_url (parser, lines[i],
-					totem_pl_parser_get_extinfo_title (extinfo, lines, i));
+			if (totem_pl_parser_parse_internal (parser, lines[i])
+					!= TOTEM_PL_PARSER_RESULT_SUCCESS) {
+				totem_pl_parser_add_one_url (parser, lines[i],
+						totem_pl_parser_get_extinfo_title (extinfo, lines, i));
+			}
 			extinfo = FALSE;
 		} else if (lines[i][0] == '\\' && lines[i][1] == '\\') {
 			/* ... Or it's in the windows smb form
@@ -1231,7 +1374,7 @@ parse_asx_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
 {
 	xmlNodePtr node;
 	TotemPlParserResult retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
-	guchar *title, *url;
+	xmlChar *title, *url;
 	char *fullpath = NULL;
 
 	title = NULL;
@@ -1244,19 +1387,19 @@ parse_asx_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
 		/* ENTRY can only have one title node but multiple REFs */
 		if (g_ascii_strcasecmp ((char *)node->name, "ref") == 0
 				|| g_ascii_strcasecmp ((char *)node->name, "entryref") == 0) {
-			unsigned char *tmp;
+			xmlChar *tmp;
 
-			tmp = xmlGetProp (node, (guchar *)"href");
+			tmp = xmlGetProp (node, (const xmlChar *)"href");
 			if (tmp == NULL)
-				tmp = xmlGetProp (node, (guchar *)"HREF");
+				tmp = xmlGetProp (node, (const xmlChar *)"HREF");
 			if (tmp == NULL)
 				continue;
-			if (url == NULL || g_str_has_prefix ((char *)tmp, "mms:") != FALSE) {
-				g_free (url);
+			/* FIXME, should we prefer mms streams, or non-mms?
+			 * See bug #352559 */
+			if (url == NULL)
 				url = tmp;
-			} else {
-				g_free (tmp);
-			}
+			else
+				xmlFree (tmp);
 
 			continue;
 		}
@@ -1266,17 +1409,14 @@ parse_asx_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
 	}
 
 	if (url == NULL) {
-		g_free (title);
+		if (title)
+			xmlFree (title);
 		return TOTEM_PL_PARSER_RESULT_ERROR;
 	}
 
-	if (strstr ((char *)url, "://") == NULL && url[0] != '/') {
-		fullpath = g_strdup_printf ("%s/%s", base, url);
-	} else {
-		fullpath = g_strdup ((char *)url);
-	}
+	fullpath = totem_pl_resolve_url (base, (char *)url);
 
-	g_free (url);
+	xmlFree (url);
 
 	/* .asx files can contain references to other .asx files */
 	retval = totem_pl_parser_parse_internal (parser, fullpath);
@@ -1286,7 +1426,8 @@ parse_asx_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
 	}
 
 	g_free (fullpath);
-	g_free (title);
+	if (title)
+		xmlFree (title);
 
 	return retval;
 }
@@ -1295,9 +1436,10 @@ static gboolean
 parse_asx_entries (TotemPlParser *parser, char *base, xmlDocPtr doc,
 		xmlNodePtr parent)
 {
-	guchar *title = NULL;
+	xmlChar *title = NULL;
 	xmlNodePtr node;
 	TotemPlParserResult retval = TOTEM_PL_PARSER_RESULT_ERROR;
+	xmlChar *newbase = NULL;
 
 	for (node = parent->children; node != NULL; node = node->next) {
 		if (node->name == NULL)
@@ -1306,6 +1448,13 @@ parse_asx_entries (TotemPlParser *parser, char *base, xmlDocPtr doc,
 		if (g_ascii_strcasecmp ((char *)node->name, "title") == 0) {
 			title = xmlNodeListGetString(doc, node->children, 1);
 		}
+		if (g_ascii_strcasecmp ((char *)node->name, "base") == 0) {
+			newbase = xmlGetProp (node, (const xmlChar *)"href");
+			if (newbase == NULL)
+				newbase = xmlGetProp (node, (const xmlChar *)"HREF");
+			if (newbase != NULL)
+				base = (char *)newbase;
+		}
 
 		if (g_ascii_strcasecmp ((char *)node->name, "entry") == 0) {
 			/* Whee found an entry here, find the REF and TITLE */
@@ -1325,7 +1474,11 @@ parse_asx_entries (TotemPlParser *parser, char *base, xmlDocPtr doc,
 		}
 	}
 
-	g_free (title);
+	if (newbase)
+		xmlFree (newbase);
+
+	if (title)
+		xmlFree (title);
 
 	return retval;
 }
@@ -1389,10 +1542,7 @@ totem_pl_parser_add_block (TotemPlParser *parser, const char *url, gpointer data
 static TotemPlParserResult
 totem_pl_parser_add_ra (TotemPlParser *parser, const char *url, gpointer data)
 {
-	if (data == NULL
-			|| (g_str_has_prefix (data, "http://") == FALSE
-			&& g_str_has_prefix (data, "rtsp://") == FALSE
-			&& g_str_has_prefix (data, "pnm://") == FALSE)) {
+	if (data == NULL || totem_pl_parser_is_ra (data, strlen (data)) == FALSE) {
 		totem_pl_parser_add_one_url (parser, url, NULL);
 		return TOTEM_PL_PARSER_RESULT_SUCCESS;
 	}
@@ -1404,26 +1554,22 @@ static gboolean
 parse_smil_video_entry (TotemPlParser *parser, char *base,
 		char *url, char *title)
 {
-	if (strstr (url, "://") != NULL || url[0] == '/') {
-		totem_pl_parser_add_one_url (parser, url, title);
-	} else {
-		char *fullpath;
+	char *fullpath;
 
-		fullpath = g_strdup_printf ("%s/%s", base, url);
-		totem_pl_parser_add_one_url (parser, fullpath, title);
+	fullpath = totem_pl_resolve_url (base, url);
+	totem_pl_parser_add_one_url (parser, fullpath, title);
 
-		g_free (fullpath);
-	}
+	g_free (fullpath);
 
 	return TRUE;
 }
 
 static gboolean
 parse_smil_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
-		xmlNodePtr parent)
+		xmlNodePtr parent, xmlChar *parent_title)
 {
 	xmlNodePtr node;
-	guchar *title, *url;
+	xmlChar *title, *url;
 	TotemPlParserResult retval = TOTEM_PL_PARSER_RESULT_ERROR;
 
 	title = NULL;
@@ -1436,20 +1582,24 @@ parse_smil_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
 
 		/* ENTRY should only have one ref and one title nodes */
 		if (g_ascii_strcasecmp ((char *)node->name, "video") == 0 || g_ascii_strcasecmp ((char *)node->name, "audio") == 0) {
-			url = xmlGetProp (node, (guchar *)"src");
-			title = xmlGetProp (node, (guchar *)"title");
+			url = xmlGetProp (node, (const xmlChar *)"src");
+			title = xmlGetProp (node, (const xmlChar *)"title");
 
 			if (url != NULL) {
 				if (parse_smil_video_entry (parser,
-						base, (char *)url, (char *)title) != FALSE)
+						base, (char *)url,
+						title ? (char *)title
+						: (char *)parent_title) != FALSE)
 					retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
 			}
 
-			g_free (title);
-			g_free (url);
+			if (title)
+				xmlFree (title);
+			if (url)
+				xmlFree (url);
 		} else {
 			if (parse_smil_entry (parser,
-						base, doc, node) != FALSE)
+						base, doc, node, parent_title) != FALSE)
 				retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
 		}
 	}
@@ -1457,12 +1607,37 @@ parse_smil_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
 	return retval;
 }
 
+static xmlChar *
+parse_smil_head (TotemPlParser *parser, xmlDocPtr doc, xmlNodePtr parent)
+{
+	xmlNodePtr node;
+	xmlChar *title = NULL;
+
+	for (node = parent->children; node != NULL; node = node->next) {
+		if (g_ascii_strcasecmp ((char *)node->name, "meta") == 0) {
+			xmlChar *prop;
+			prop = xmlGetProp (node, (const xmlChar *)"name");
+			if (prop != NULL && g_ascii_strcasecmp ((char *)prop, "title") == 0) {
+				title = xmlGetProp (node, (const xmlChar *)"content");
+				if (title != NULL) {
+					xmlFree (prop);
+					break;
+				}
+			}
+			xmlFree (prop);
+		}
+	}
+
+	return title;
+}
+
 static gboolean
 parse_smil_entries (TotemPlParser *parser, char *base, xmlDocPtr doc,
 		xmlNodePtr parent)
 {
 	xmlNodePtr node;
 	TotemPlParserResult retval = TOTEM_PL_PARSER_RESULT_ERROR;
+	xmlChar *title = NULL;
 
 	for (node = parent->children; node != NULL; node = node->next) {
 		if (node->name == NULL)
@@ -1470,12 +1645,17 @@ parse_smil_entries (TotemPlParser *parser, char *base, xmlDocPtr doc,
 
 		if (g_ascii_strcasecmp ((char *)node->name, "body") == 0) {
 			if (parse_smil_entry (parser, base,
-						doc, node) != FALSE)
+						doc, node, title) != FALSE)
 				retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
+		} else if (title == NULL) {
+			if (g_ascii_strcasecmp ((char *)node->name, "head") == 0)
+				title = parse_smil_head (parser, doc, node);
 		}
-
 	}
 
+	if (title != NULL)
+		xmlFree (title);
+
 	return retval;
 }
 
@@ -1527,11 +1707,7 @@ totem_pl_parser_add_asf (TotemPlParser *parser, const char *url, gpointer data)
 		return TOTEM_PL_PARSER_RESULT_SUCCESS;
 	}
 
-	if (g_str_has_prefix (data, "[Reference]") == FALSE
-		 && g_ascii_strncasecmp (data, "<ASX", strlen ("<ASX")) != 0
-		 && strstr (data, "<ASX") == NULL
-		 && strstr (data, "<asx") == NULL
-		 && g_str_has_prefix (data, "ASF ") == FALSE) {
+	if (totem_pl_parser_is_asf (data, strlen (data)) == FALSE) {
 		totem_pl_parser_add_one_url (parser, url, NULL);
 		return TOTEM_PL_PARSER_RESULT_SUCCESS;
 	}
@@ -1544,7 +1720,8 @@ totem_pl_parser_add_quicktime_metalink (TotemPlParser *parser, const char *url,
 {
 	xmlDocPtr doc;
 	xmlNodePtr node;
-	char *contents = NULL, *src;
+	char *contents = NULL;
+	xmlChar *src;
 	int size;
 
 	contents = totem_pl_parser_read_entire_file (url, &size);
@@ -1579,14 +1756,15 @@ totem_pl_parser_add_quicktime_metalink (TotemPlParser *parser, const char *url,
 		return TOTEM_PL_PARSER_RESULT_ERROR;
 	}
 
-	src = (char *) xmlGetProp (node, (guchar *)"src");
+	src = xmlGetProp (node, (const xmlChar *)"src");
 	if (!src) {
 		xmlFreeDoc (doc);
 		return TOTEM_PL_PARSER_RESULT_ERROR;
 	}
 
-	totem_pl_parser_add_one_url (parser, src, NULL);
+	totem_pl_parser_add_one_url (parser, (char *) src, NULL);
 
+	xmlFree (src);
 	xmlFreeDoc (doc);
 
 	return TOTEM_PL_PARSER_RESULT_SUCCESS;
@@ -1595,7 +1773,7 @@ totem_pl_parser_add_quicktime_metalink (TotemPlParser *parser, const char *url,
 static TotemPlParserResult
 totem_pl_parser_add_quicktime (TotemPlParser *parser, const char *url, gpointer data)
 {
-	if (data == NULL || strstr (data, "<?quicktime") == NULL) {
+	if (data == NULL || totem_pl_parser_is_quicktime (data, strlen (data)) == FALSE) {
 		totem_pl_parser_add_one_url (parser, url, NULL);
 		return TOTEM_PL_PARSER_RESULT_SUCCESS;
 	}
@@ -1659,12 +1837,82 @@ totem_pl_parser_dir_compare (GnomeVFSFileInfo *a, GnomeVFSFileInfo *b)
 	}
 }
 
+static char *
+totem_pl_parser_iso_get_title (const char *url)
+{
+	char *fname;
+	FILE  *file;
+#define BUFFER_SIZE 128
+	char buf [BUFFER_SIZE+1];
+	int res;
+	char *str;
+
+	fname = g_filename_from_uri (url, NULL, NULL);
+	if (fname == NULL)
+		return NULL;
+
+	file = fopen (fname, "rb");
+	if (file == NULL)
+		return NULL;
+
+	/* Verify we have an ISO image */
+	/* This check is for the raw sector images */
+	res = fseek (file, 37633L, SEEK_SET);
+	if (res != 0) {
+		fclose (file);
+		return NULL;
+	}
+
+	res = fread (buf, sizeof (char), 5, file);
+	if (res != 5 || strncmp (buf, "CD001", 5) != 0) {
+		/* Standard ISO images */
+		res = fseek (file, 32769L, SEEK_SET);
+		if (res != 0) {
+			fclose (file);
+			return NULL;
+		}
+		res = fread (buf, sizeof (char), 5, file);
+		if (res != 5 || strncmp (buf, "CD001", 5) != 0) {
+			/* High Sierra images */
+			res = fseek (file, 32776L, SEEK_SET);
+			if (res != 0) {
+				fclose (file);
+				return NULL;
+			}
+			res = fread (buf, sizeof (char), 5, file);
+			if (res != 5 || strncmp (buf, "CDROM", 5) != 0) {
+				fclose (file);
+				return NULL;
+			}
+		}
+	}
+	/* Extract the volume label from the image */
+	res = fseek (file, 32808L, SEEK_SET);
+	if (res != 0) {
+		fclose (file);
+		return NULL;
+	}
+	res = fread (buf, sizeof(char), BUFFER_SIZE, file);
+	fclose (file);
+	if (res != BUFFER_SIZE)
+		return NULL;
+
+	buf [BUFFER_SIZE] = '\0';
+	str = g_strdup (g_strstrip (buf));
+	if (!g_utf8_validate (str, -1, NULL)) {
+		g_free (str);
+		return NULL;
+	}
+
+	return str;
+}
+
 static TotemPlParserResult
 totem_pl_parser_add_iso (TotemPlParser *parser, const char *url,
 		gpointer data)
 {
 	GnomeVFSFileInfo *info;
-	char *item;
+	char *item, *label;
 
 	/* This is a hack, it could be a VCD or DVD */
 	if (g_str_has_prefix (url, "file://") == FALSE)
@@ -1683,7 +1931,10 @@ totem_pl_parser_add_iso (TotemPlParser *parser, const char *url,
 	}
 
 	gnome_vfs_file_info_unref (info);
-	totem_pl_parser_add_one_url (parser, item, NULL);
+
+	label = totem_pl_parser_iso_get_title (url);
+	totem_pl_parser_add_one_url (parser, item, label);
+	g_free (label);
 	g_free (item);
 
 	return TOTEM_PL_PARSER_RESULT_SUCCESS;
@@ -1702,6 +1953,128 @@ totem_pl_parser_add_cue (TotemPlParser *parser, const char *url,
 	return TOTEM_PL_PARSER_RESULT_SUCCESS;
 }
 
+static gboolean
+parse_xspf_track (TotemPlParser *parser, char *base, xmlDocPtr doc,
+		xmlNodePtr parent)
+{
+	xmlNodePtr node;
+	xmlChar *title, *url;
+	gchar *fullpath;
+	TotemPlParserResult retval = TOTEM_PL_PARSER_RESULT_ERROR;
+	
+	title = NULL;
+	url = NULL;
+
+	for (node = parent->children; node != NULL; node = node->next)
+	{
+		if (node->name == NULL)
+			continue;
+
+		if (g_ascii_strcasecmp ((char *)node->name, "location") == 0)
+			url = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
+
+		if (g_ascii_strcasecmp ((char *)node->name, "title") == 0)
+			title = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
+	}
+
+	if (url == NULL) {
+		if (title)
+			xmlFree (title);
+		return TOTEM_PL_PARSER_RESULT_ERROR;
+	}
+
+	fullpath = totem_pl_resolve_url (base, (char *)url);
+	totem_pl_parser_add_one_url (parser, fullpath, (char *)title);
+	retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
+
+	if (title)
+		xmlFree (title);
+	if (url)
+		xmlFree (url);
+	g_free (fullpath);
+
+	return retval;
+}
+
+static gboolean
+parse_xspf_trackList (TotemPlParser *parser, char *base, xmlDocPtr doc,
+		xmlNodePtr parent)
+{
+	xmlNodePtr node;
+	TotemPlParserResult retval = TOTEM_PL_PARSER_RESULT_ERROR;
+
+	for (node = parent->children; node != NULL; node = node->next)
+	{
+		if (node->name == NULL)
+			continue;
+
+		if (g_ascii_strcasecmp ((char *)node->name, "track") == 0)
+			if (parse_xspf_track (parser, base, doc, node) != FALSE)
+				retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
+	}
+
+	return retval;
+}
+
+static gboolean
+parse_xspf_entries (TotemPlParser *parser, char *base, xmlDocPtr doc,
+		xmlNodePtr parent)
+{
+	xmlNodePtr node;
+	TotemPlParserResult retval = TOTEM_PL_PARSER_RESULT_ERROR;
+
+	for (node = parent->children; node != NULL; node = node->next) {
+		if (node->name == NULL)
+			continue;
+
+		if (g_ascii_strcasecmp ((char *)node->name, "trackList") == 0)
+			if (parse_xspf_trackList (parser, base, doc, node) != FALSE)
+				retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
+	}
+
+	return retval;
+}
+
+static TotemPlParserResult
+totem_pl_parser_add_xspf (TotemPlParser *parser, const char *url,
+		gpointer data)
+{
+	xmlDocPtr doc;
+	xmlNodePtr node;
+	char *contents = NULL, *base;
+	int size;
+	gboolean retval = TOTEM_PL_PARSER_RESULT_UNHANDLED;
+
+	contents = totem_pl_parser_read_entire_file (url, &size);
+	if (contents == NULL)
+		return FALSE;
+
+	doc = xmlParseMemory (contents, size);
+	if (doc == NULL)
+		doc = xmlRecoverMemory (contents, size);
+	g_free (contents);
+
+	/* If the document has no root, or no name */
+	if(!doc || !doc->children
+			|| !doc->children->name
+			|| g_ascii_strcasecmp ((char *)doc->children->name,
+				"playlist") != 0) {
+		if (doc != NULL)
+			xmlFreeDoc(doc);
+		return TOTEM_PL_PARSER_RESULT_ERROR;
+	}
+
+	base = totem_pl_parser_base_url (url);
+
+	for (node = doc->children; node != NULL; node = node->next)
+		if (parse_xspf_entries (parser, base, doc, node) != FALSE)
+			retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
+
+	g_free (base);
+	xmlFreeDoc(doc);
+	return retval;
+}
+
 static TotemPlParserResult
 totem_pl_parser_add_directory (TotemPlParser *parser, const char *url,
 			   gpointer data)
@@ -1709,17 +2082,22 @@ totem_pl_parser_add_directory (TotemPlParser *parser, const char *url,
 	MediaType type;
 	GList *list, *l;
 	GnomeVFSResult res;
+	char *media_url;
 
-	if (parser->priv->recurse_level == 1) {
-		char *media_url;
+	type = totem_cd_detect_type_from_dir (url, &media_url, NULL);
+	if (type != MEDIA_TYPE_DATA && type != MEDIA_TYPE_ERROR) {
+		if (media_url != NULL) {
+			char *basename = NULL, *fname;
 
-		type = totem_cd_detect_type_from_dir (url, &media_url, NULL);
-		if (type != MEDIA_TYPE_DATA && type != MEDIA_TYPE_ERROR) {
-			if (media_url != NULL) {
-				totem_pl_parser_add_one_url (parser, media_url, NULL);
-				g_free (media_url);
-				return TOTEM_PL_PARSER_RESULT_SUCCESS;
+			fname = g_filename_from_uri (url, NULL, NULL);
+			if (fname != NULL) {
+				basename = g_filename_display_basename (fname);
+				g_free (fname);
 			}
+			totem_pl_parser_add_one_url (parser, media_url, basename);
+			g_free (basename);
+			g_free (media_url);
+			return TOTEM_PL_PARSER_RESULT_SUCCESS;
 		}
 	}
 
@@ -1759,43 +2137,134 @@ totem_pl_parser_add_directory (TotemPlParser *parser, const char *url,
 	return TOTEM_PL_PARSER_RESULT_SUCCESS;
 }
 
+#endif /* !TOTEM_PL_PARSER_MINI */
+
+static gboolean
+totem_pl_parser_is_ra (const char *data, gsize len)
+{
+	/* pnm is the smallest, 6 chars */
+	if (len < 7)
+		return FALSE;
+
+	if (g_str_has_prefix (data, "http://") != FALSE
+			|| g_str_has_prefix (data, "rtsp://") != FALSE
+			|| g_str_has_prefix (data, "pnm://") != FALSE) {
+		return TRUE;
+	}
+	return FALSE;
+}
+
+static gboolean
+totem_pl_parser_is_asf (const char *data, gsize len)
+{
+	char *buffer;
+
+	if (len == 0)
+		return FALSE;
+
+	if (g_str_has_prefix (data, "[Reference]") != FALSE
+			|| g_ascii_strncasecmp (data, "<ASX", strlen ("<ASX")) == 0
+			|| g_str_has_prefix (data, "ASF ") != FALSE) {
+		return TRUE;
+	}
+
+	if (len > MIME_READ_CHUNK_SIZE)
+		len = MIME_READ_CHUNK_SIZE;
+
+	/* FIXME would be nicer to have an strnstr */
+	buffer = g_memdup (data, len);
+	if (buffer == NULL) {
+		g_warning ("Couldn't dup data in totem_pl_parser_is_asf");
+		return FALSE;
+	}
+	buffer[len - 1] = '\0';
+	if (strstr (buffer, "<ASX") != NULL
+			|| strstr (buffer, "<asx") != NULL) {
+		g_free (buffer);
+		return TRUE;
+	}
+	g_free (buffer);
+	return FALSE;
+}
+
+static gboolean
+totem_pl_parser_is_quicktime (const char *data, gsize len)
+{
+	char *buffer;
+
+	if (len == 0)
+		return FALSE;
+	if (len > MIME_READ_CHUNK_SIZE)
+		len = MIME_READ_CHUNK_SIZE;
+
+	/* FIXME would be nicer to have an strnstr */
+	buffer = g_memdup (data, len);
+	if (buffer == NULL) {
+		g_warning ("Couldn't dup data in totem_pl_parser_is_quicktime");
+		return FALSE;
+	}
+	buffer[len - 1] = '\0';
+	if (strstr (buffer, "<?quicktime") != NULL) {
+		g_free (buffer);
+		return TRUE;
+	}
+	g_free (buffer);
+	return FALSE;
+}
+
+#ifndef TOTEM_PL_PARSER_MINI
+#define PLAYLIST_TYPE(mime,cb,identcb) { mime, cb, identcb }
+#define PLAYLIST_TYPE2(mime,cb,identcb) { mime, cb, identcb }
+#else
+#define PLAYLIST_TYPE(mime,cb,identcb) { mime }
+#define PLAYLIST_TYPE2(mime,cb,identcb) { mime, identcb }
+#endif
+
 /* These ones need a special treatment, mostly parser formats */
 static PlaylistTypes special_types[] = {
-	{ "audio/x-mpegurl", totem_pl_parser_add_m3u },
-	{ "audio/playlist", totem_pl_parser_add_m3u },
-	{ "audio/x-ms-asx", totem_pl_parser_add_asx },
-	{ "audio/x-scpls", totem_pl_parser_add_pls },
-	{ "application/x-smil", totem_pl_parser_add_smil },
-	{ "application/smil", totem_pl_parser_add_smil },
-	{ "application/x-gnome-app-info", totem_pl_parser_add_desktop },
-	{ "application/x-desktop", totem_pl_parser_add_desktop },
-	{ "x-directory/normal", totem_pl_parser_add_directory },
-	{ "video/x-ms-wvx", totem_pl_parser_add_asx },
-	{ "audio/x-ms-wax", totem_pl_parser_add_asx },
-	{ "application/x-cd-image", totem_pl_parser_add_iso },
-	{ "application/x-cue", totem_pl_parser_add_cue },
-	{ "x-special/device-block", totem_pl_parser_add_block },
-};
-
-static PlaylistTypes ignore_types[] = {
-	{ "image/*", NULL },
-	{ "text/plain", NULL },
-	{ "application/x-rar", NULL },
-	{ "application/zip", NULL },
-	{ "application/x-trash", NULL },
+	PLAYLIST_TYPE ("audio/x-mpegurl", totem_pl_parser_add_m3u, NULL),
+	PLAYLIST_TYPE ("audio/playlist", totem_pl_parser_add_m3u, NULL),
+	PLAYLIST_TYPE ("audio/x-ms-asx", totem_pl_parser_add_asx, NULL),
+	PLAYLIST_TYPE ("audio/x-scpls", totem_pl_parser_add_pls, NULL),
+	PLAYLIST_TYPE ("application/x-smil", totem_pl_parser_add_smil, NULL),
+	PLAYLIST_TYPE ("application/smil", totem_pl_parser_add_smil, NULL),
+	PLAYLIST_TYPE ("video/x-ms-wvx", totem_pl_parser_add_asx, NULL),
+	PLAYLIST_TYPE ("audio/x-ms-wax", totem_pl_parser_add_asx, NULL),
+	PLAYLIST_TYPE ("application/xspf+xml", totem_pl_parser_add_xspf, NULL),
+#ifndef TOTEM_PL_PARSER_MINI
+	PLAYLIST_TYPE ("application/x-desktop", totem_pl_parser_add_desktop, NULL),
+	PLAYLIST_TYPE ("application/x-gnome-app-info", totem_pl_parser_add_desktop, NULL),
+	PLAYLIST_TYPE ("application/x-cd-image", totem_pl_parser_add_iso, NULL),
+	PLAYLIST_TYPE ("application/x-cue", totem_pl_parser_add_cue, NULL),
+	PLAYLIST_TYPE ("x-directory/normal", totem_pl_parser_add_directory, NULL),
+	PLAYLIST_TYPE ("x-special/device-block", totem_pl_parser_add_block, NULL),
+#endif
 };
 
 /* These ones are "dual" types, might be a video, might be a parser */
 static PlaylistTypes dual_types[] = {
-	{ "audio/x-real-audio", totem_pl_parser_add_ra },
-	{ "audio/x-pn-realaudio", totem_pl_parser_add_ra },
-	{ "application/vnd.rn-realmedia", totem_pl_parser_add_ra },
-	{ "audio/x-pn-realaudio-plugin", totem_pl_parser_add_ra },
-	{ "audio/vnd.rn-realaudio", totem_pl_parser_add_ra },
-	{ "text/plain", totem_pl_parser_add_ra },
-	{ "video/x-ms-asf", totem_pl_parser_add_asf },
-	{ "video/x-ms-wmv", totem_pl_parser_add_asf },
-	{ "video/quicktime", totem_pl_parser_add_quicktime },
+	PLAYLIST_TYPE2 ("audio/x-real-audio", totem_pl_parser_add_ra, totem_pl_parser_is_ra),
+	PLAYLIST_TYPE2 ("audio/x-pn-realaudio", totem_pl_parser_add_ra, totem_pl_parser_is_ra),
+	PLAYLIST_TYPE2 ("application/ram", totem_pl_parser_add_ra, totem_pl_parser_is_ra),
+	PLAYLIST_TYPE2 ("application/vnd.rn-realmedia", totem_pl_parser_add_ra, totem_pl_parser_is_ra),
+	PLAYLIST_TYPE2 ("audio/x-pn-realaudio-plugin", totem_pl_parser_add_ra, totem_pl_parser_is_ra),
+	PLAYLIST_TYPE2 ("audio/vnd.rn-realaudio", totem_pl_parser_add_ra, totem_pl_parser_is_ra),
+	PLAYLIST_TYPE2 ("audio/x-realaudio", totem_pl_parser_add_ra, totem_pl_parser_is_ra),
+	PLAYLIST_TYPE2 ("text/plain", totem_pl_parser_add_ra, totem_pl_parser_is_ra),
+	PLAYLIST_TYPE2 ("video/x-ms-asf", totem_pl_parser_add_asf, totem_pl_parser_is_asf),
+	PLAYLIST_TYPE2 ("video/x-ms-wmv", totem_pl_parser_add_asf, totem_pl_parser_is_asf),
+	PLAYLIST_TYPE2 ("video/quicktime", totem_pl_parser_add_quicktime, totem_pl_parser_is_quicktime),
+	PLAYLIST_TYPE2 ("application/x-quicktime-media-link", totem_pl_parser_add_quicktime, totem_pl_parser_is_quicktime),
+};
+
+#ifndef TOTEM_PL_PARSER_MINI
+
+static PlaylistTypes ignore_types[] = {
+	PLAYLIST_TYPE ("image/*", NULL, NULL),
+	PLAYLIST_TYPE ("text/plain", NULL, NULL),
+	PLAYLIST_TYPE ("application/x-rar", NULL, NULL),
+	PLAYLIST_TYPE ("application/zip", NULL, NULL),
+	PLAYLIST_TYPE ("application/x-trash", NULL, NULL),
 };
 
 static gboolean
@@ -1846,7 +2315,7 @@ totem_pl_parser_ignore (TotemPlParser *parser, const char *url)
 		return TRUE;
 
 	mimetype = gnome_vfs_get_file_mime_type (url, NULL, TRUE);
-	if (mimetype == NULL || strcmp (mimetype, "application/octet-stream") == 0)
+	if (mimetype == NULL || strcmp (mimetype, GNOME_VFS_MIME_TYPE_UNKNOWN) == 0)
 		return FALSE;
 
 	for (i = 0; i < G_N_ELEMENTS (special_types); i++)
@@ -1867,6 +2336,34 @@ totem_pl_parser_ignore (TotemPlParser *parser, const char *url)
 	return TRUE;
 }
 
+static gboolean
+totem_pl_parser_ignore_from_mimetype (TotemPlParser *parser, const char *mimetype)
+{
+	char *super;
+	guint i;
+
+	super = gnome_vfs_get_supertype_from_mime_type (mimetype);
+	for (i = 0; i < G_N_ELEMENTS (ignore_types) && super != NULL; i++) {
+		if (gnome_vfs_mime_type_is_supertype (ignore_types[i].mimetype) != FALSE) {
+			if (strcmp (super, ignore_types[i].mimetype) == 0) {
+				g_free (super);
+				return TRUE;
+			}
+		} else {
+			GnomeVFSMimeEquivalence eq;
+
+			eq = gnome_vfs_mime_type_get_equivalence (mimetype, ignore_types[i].mimetype);
+			if (eq == GNOME_VFS_MIME_PARENT || eq == GNOME_VFS_MIME_IDENTICAL) {
+				g_free (super);
+				return TRUE;
+			}
+		}
+	}
+	g_free (super);
+
+	return FALSE;
+}
+
 static TotemPlParserResult
 totem_pl_parser_parse_internal (TotemPlParser *parser, const char *url)
 {
@@ -1874,7 +2371,6 @@ totem_pl_parser_parse_internal (TotemPlParser *parser, const char *url)
 	guint i;
 	gpointer data = NULL;
 	TotemPlParserResult ret = TOTEM_PL_PARSER_RESULT_ERROR;
-	char *super;
 	gboolean found = FALSE;
 
 	if (parser->priv->recurse_level > RECURSE_LEVEL_MAX)
@@ -1939,24 +2435,8 @@ totem_pl_parser_parse_internal (TotemPlParser *parser, const char *url)
 	if (ret == TOTEM_PL_PARSER_RESULT_SUCCESS)
 		return ret;
 
-	super = gnome_vfs_get_supertype_from_mime_type (mimetype);
-	for (i = 0; i < G_N_ELEMENTS (ignore_types) && super != NULL && found == FALSE; i++) {
-		if (gnome_vfs_mime_type_is_supertype (ignore_types[i].mimetype) != FALSE) {
-			if (strcmp (super, ignore_types[i].mimetype) == 0) {
-				g_free (super);
-				return TOTEM_PL_PARSER_RESULT_IGNORED;
-			}
-		} else {
-			GnomeVFSMimeEquivalence eq;
-
-			eq = gnome_vfs_mime_type_get_equivalence (mimetype, ignore_types[i].mimetype);
-			if (eq == GNOME_VFS_MIME_PARENT || eq == GNOME_VFS_MIME_IDENTICAL) {
-				g_free (super);
-				return TOTEM_PL_PARSER_RESULT_IGNORED;
-			}
-		}
-	}
-	g_free (super);
+	if (totem_pl_parser_ignore_from_mimetype (parser, mimetype))
+		return TOTEM_PL_PARSER_RESULT_IGNORED;
 
 	if (ret != TOTEM_PL_PARSER_RESULT_SUCCESS && parser->priv->fallback) {
 		totem_pl_parser_add_one_url (parser, url, NULL);
@@ -1983,7 +2463,7 @@ totem_pl_parser_parse (TotemPlParser *parser, const char *url,
 			TOTEM_PL_PARSER_RESULT_IGNORED);
 
 	parser->priv->recurse_level = 0;
-	parser->priv->fallback = fallback;
+	parser->priv->fallback = fallback != FALSE;
 	return totem_pl_parser_parse_internal (parser, url);
 }
 
@@ -2007,3 +2487,65 @@ totem_pl_parser_add_ignored_mimetype (TotemPlParser *parser,
 		(parser->priv->ignore_mimetypes, g_strdup (mimetype));
 }
 
+#endif /* !TOTEM_PL_PARSER_MINI */
+
+#define D(x) if (debug) x
+
+gboolean
+totem_pl_parser_can_parse_from_data (const char *data,
+				     gsize len,
+				     gboolean debug)
+{
+	const char *mimetype;
+	guint i;
+
+	g_return_val_if_fail (data != NULL, FALSE);
+
+	/* Bad cast! */
+	mimetype = gnome_vfs_get_mime_type_for_data ((gpointer) data, (int) len);
+
+	if (mimetype == NULL || strcmp (GNOME_VFS_MIME_TYPE_UNKNOWN, mimetype) == 0) {
+		D(g_message ("totem_pl_parser_can_parse_from_data couldn't get mimetype"));
+		return FALSE;
+	}
+
+	for (i = 0; i < G_N_ELEMENTS(special_types); i++) {
+		if (strcmp (special_types[i].mimetype, mimetype) == 0) {
+			D(g_message ("Is special type '%s'\n", mimetype));
+			return TRUE;
+		}
+	}
+
+	for (i = 0; i < G_N_ELEMENTS(dual_types); i++) {
+		if (strcmp (dual_types[i].mimetype, mimetype) == 0) {
+			D(g_message ("Is dual type '%s'\n", mimetype));
+			if (dual_types[i].iden != NULL)
+				return (* dual_types[i].iden) (data, len);
+			return FALSE;
+		}
+	}
+
+	return FALSE;
+}
+
+gboolean
+totem_pl_parser_can_parse_from_filename (const char *filename, gboolean debug)
+{
+	char *contents;
+	gsize len;
+	gboolean retval;
+
+	g_return_val_if_fail (filename != NULL, FALSE);
+
+	/* FIXME probably a good idea to only get the first few chunks */
+	if (g_file_get_contents (filename, &contents, &len, NULL) == FALSE) {
+		D(g_message ("couldn't get the contents of %s", filename));
+		return FALSE;
+	}
+
+	retval = totem_pl_parser_can_parse_from_data (contents, len, debug);
+	g_free (contents);
+
+	return retval;
+}
+
