@@ -77,14 +77,14 @@ totem_lang_table_parse_start_tag (GMarkupParseContext *ctx,
 			/* skip if empty */
 			if (**attr_values)
 			{
-				g_return_if_fail (strlen (*attr_values) == 3);
+				g_return_if_fail (strlen (*attr_values) == 3 || strcmp (*attr_values, "qaa-qtz") == 0);
 				ccode_longB = *attr_values;
 			}
 		} else if (g_str_equal (*attr_names, "iso_639_2T_code")) {
 			/* skip if empty */
 			if (**attr_values)
 			{
-				g_return_if_fail (strlen (*attr_values) == 3);
+				g_return_if_fail (strlen (*attr_values) == 3 || strcmp (*attr_values, "qaa-qtz") == 0);
 				ccode_longT = *attr_values;
 			}
 		} else if (g_str_equal (*attr_names, "name")) {
@@ -334,7 +334,7 @@ totem_sublang_equal_lists (GList *orig, GList *new)
 static void
 totem_languages_update (Totem *totem, GList *list)
 {
-	GtkWidget *item, *submenu;
+	GtkWidget *item, *submenu, *old_submenu;
 	GtkWidget *lang_menu;
 	int current;
 
@@ -348,6 +348,11 @@ totem_languages_update (Totem *totem, GList *list)
 
 	/* Languages */
 	item = glade_xml_get_widget (totem->xml, "tmw_languages_menu_item");
+
+	old_submenu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (item));
+	if (old_submenu != NULL)
+		gtk_menu_popdown (GTK_MENU (old_submenu));
+
 	submenu = glade_xml_get_widget (totem->xml, "tmw_menu_languages");
 	if (lang_menu == NULL)
 	{
@@ -365,7 +370,7 @@ totem_languages_update (Totem *totem, GList *list)
 static void
 totem_subtitles_update (Totem *totem, GList *list)
 {
-	GtkWidget *item, *submenu;
+	GtkWidget *item, *submenu, *old_submenu;
 	GtkWidget *sub_menu;
 	int current;
 
@@ -379,6 +384,11 @@ totem_subtitles_update (Totem *totem, GList *list)
 
 	/* Subtitles */
 	item = glade_xml_get_widget (totem->xml, "tmw_subtitles_menu_item");
+
+	old_submenu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (item));
+	if (old_submenu != NULL)
+		gtk_menu_popdown (GTK_MENU (old_submenu));
+
 	submenu = glade_xml_get_widget (totem->xml, "tmw_menu_subtitles");
 	if (sub_menu == NULL)
 	{
@@ -591,6 +601,7 @@ add_device_to_menu (GObject *device, GtkMenu *menu, gint position, Totem *totem)
 		&my_gnome_vfs_volume_get_mount_path,
 		&gnome_vfs_drive_get_device_path);
 
+	g_strchomp (name);
 	label = g_strdup_printf (_("Play Disc '%s'"), name);
 	g_free (name);
 	menu_item = gtk_image_menu_item_new_with_label (label);
